#include "OpcodeBase.hpp"
#include <m_pd.h>
#include <string>
#include <vector>
#include <boost/tokenizer.hpp>

// If this is possible at all: 
// Call sys_main in a separate thread to start PD.
// Before each tick, write audio using adc_dsp, or perhaps using send.
// Before each tick, enqueue MIDI using sys_midibytein.
// After each tick, read audio from sys_soundout if it can be accessed,
// perhaps using recv,
// or create a new external to do that.
// Schedule ticks using thread locking calls: sys_lock before PD does it,
// then do stuff, sys_unlock, sys_lock, etc.

class PDPatch : public OpcodeBase<PDPatch>
{
public:
    // Outputs.
    MYFLT *ipd;
    // Inputs.
    MYFLT *ipdcommand;
    // State.
    std::string pdcommand;
    std::vector<char *> tokens;
    boost::tokenizer tokenizer;
    pthread_t threadId;
    float *sys_soundin_;
    float *sys_soundout_;
    void initialize()
    {
 		if (*ipdcommand == SSTRCOD) {
			pdcommand = p->STRARG);          
		}
        tokenizer = pdcommand;
        for(tokenizer<>::iterator it = tokenizer.begin(); it != tokenizer.end(); ++it) {
            tokens.push_back(&*it);
        }
        *ipd = (MYFLT) this;
    }
    void deinitialize()
    {
        sys_exit();
    }
    void startThread()
    {
        threadId = pthread_create(0, 0, threadRoutine, this);
    }
    static void threadRoutine(void *userdata)
    {
        sys_main(tokens.size(), &tokens.front());
    }
};

class PDMidi : public OpcodeBase<PDMidi>
{
public:
    // Inputs.
    MYFLT *ipdpatch;
    MYFLT *kinstatus;
    MYFLT *kinchannel;
    MYFLT *kindata1;
    MYFLT *kindata2;
    // No outputs.
    // State.
    MYFLT oldkinstatus;
    MYFLT oldkinchannel;
    MYFLT oldkindata1;
    MYFLT oldkindata2;
    PDPatch *pdpatch;
    void initialize()
    {
        pdpatch = (PDPatch *)*ipdpatch;
        oldkinstatus = 0;
        oldkinchannel = 0;
        oldkindata1 = 0;
        oldkindata2 = 0;
    }
    void kontrol()
    {
    }
};

class PDNote : public OpcodeBase<PDNote>
{
public:
    // Inputs.
    MYFLT *ipdpatch;
    MYFLT *kinchannel;
    MYFLT *kinkey;
    MYFLT *kinvelocity;
    MYFLT *kinduration;
    // No outputs.
    // State.
    PDPatch *pdpatch;
    MYFLT oldkinchannel;
    MYFLT oldkinkey;
    MYFLT oldkinvelocity;
    MYFLT oldkinduration;
    size_t framesRemaining;
    void initialize()
    {
        pdpatch = (PDPatch *)*ipdpatch;
        oldkinchannel = 0;
        oldkinkey = 0;
        oldkinvelocity = 0;
        oldkinduration = 0;
    }
    void kontrol()
    {
        sys_lock();
        sys_unlock();
    }
};

class PDAin : public OpcodeBase<PDAin>
{
public:
    // No outputs.
    // Inputs.
    MYFLT *ipdpatch;
    MYFLT *ain1;
    MYFLT *ain2;
    // State.
    PDPatch *pdpatch;
    std::vector<MYFLT> input1;
    std::vector<MYFLT> input2;
    void initialize()
    {
        pdpatch = (PDPatch *)*ipdpatch;
    }
    void audio()
    {
    }
};

class PDAout : public OpcodeBase<PDAout>
{
public:
    // Outputs.
    MYFLT *aout1;
    MYFLT *aout2;
    // Inputs.
    MYFLT *ipdpatch;
    // State.
    PDPatch *pdpatch;
    void initialize()
    {
        pdpatch = (PDPatch *)*ipdpatch;
    }
    void audio()
    {
    }
};

OENTRY oentries[] = {
    {"pdpatch", sizeof(PDPatch), 1, "i", "S", &PDPatch::initialize_, 0, 0, &PDPatch::deinitialize_},
    {"pdmidi", sizeof(PDMidi), 5, "", "ikkjj", &PDMidi::initialize_, &PDMidi::kontrol_0, 0},
    {"pdnote", sizeof(PDNote), 5, "", "ikkkk", &PDNote::initialize_, &PDNote::kontrol_0, 0},
    {"pdain", sizeof(PDAin), 5, "", "iaa", &PDKin::initialize_, 0, &PDPatch::audio_, 0},
    {"pdaout", sizeof(PDAout), 5, "aa", "i", &PDKin::initialize_, 0, &PDPatch::audio_, 0},
};

PUBLIC int opcode_size()
{
    return sizeof(oentries);
}

PUBLIC OENTRY *opcode_init(ENVIRON *csound)
{
    return &oentries;
}
